import * as d3 from "d3";
import { Config } from "../model/config";
import { Chart } from "../model/chart";


import {
    axisBottom  as d3_axisBottom,
    axisLeft    as d3_axisLeft,
    scaleLinear as d3_scaleLinear,
    select      as d3_select,
    scaleRadial as d3_scaleRadial
  } from 'd3';

export class BasicChartService {
    getBasicChart(document: Document, config: Config): Chart {
        const body = document.body;
        const basicChart: Chart = this.generateBasicChartObject();
        basicChart.body = body;
        basicChart.document = document;

        let yMargin = this.getYAxisWidth(basicChart, config);

        let xMarginBottom = this.getXAxisHeight(basicChart, config);

        config.chart.margin.left = Math.max(config.chart.margin.left, yMargin + config.chart.fontSize * 1.5)

        config.chart.margin.bottom = Math.max(config.chart.margin.bottom, xMarginBottom + config.chart.fontSize * 1.5)



        let margin = {top: config.chart.margin.top, right: config.chart.margin.right, 
            bottom: config.chart.margin.bottom, left: config.chart.margin.left},
            width = config.chart.width - margin.left - margin.right,
            height = config.chart.height - margin.top - margin.bottom;

        

        const svG = d3.select(body).select('#chart')
            .append("svg")
                .attr("width", width + margin.left + margin.right)
                .attr("height", height + margin.top + margin.bottom)
                .style("overflow", "visible")
            .append("g")
                .attr("transform",
                    "translate(" + margin.left + "," + margin.top + ")");

        basicChart['svg'] = svG;

        
        /* Achsenformate prüfen und im zweifel überschreiben? */

        if (config.chart.type === 'radar') {
            return this.getBasicRadarChart(basicChart, config, width, height)
        } else {
            return this.getBasicStandardChart(basicChart, config)
        }
    }

    


    private getBasicRadarChart(basicChart: Chart, config: Config, width: number, height: number): Chart {

        let features = ["A","B","C","D","E","F"];

        const svG = basicChart.svg;

        let scaleSize = Math.min(width / 2, height / 2);

        let radialScale = d3.scaleLinear()
            .domain([0,10])
            .range([0, scaleSize]);

        config.chart.radarConfig.radialScale = radialScale;
        
        let ticks = [2,4,6,8,10];
        ticks.forEach(t =>
            svG.append("circle")
            .attr("cx", config.chart.width / 2)
            .attr("cy", config.chart.height / 2)
            .attr("fill", "none")
            .attr("stroke", "gray")
            .attr("r", radialScale(t))
        );


        ticks.forEach(t =>
            svG.append("text")
            .attr("x", (config.chart.width / 2) + 10)
            .attr("y", config.chart.height / 2 - radialScale(t))
            .text(t.toString())
        );

        for (var i = 0; i < features.length; i++) {
            let ft_name = features[i];
            let angle = (Math.PI / 2) + (2 * Math.PI * i / features.length);
            let line_coordinate = this.angleToCoordinate(angle, 10, radialScale, config);
            let label_coordinate = this.angleToCoordinate(angle, 10.5, radialScale, config);
        
            //draw axis line
            svG.append("line")
            .attr("x1", config.chart.width / 2)
            .attr("y1", config.chart.height / 2)
            .attr("x2", line_coordinate.x)
            .attr("y2", line_coordinate.y)
            .attr("stroke","black");
        
            //draw axis label
            svG.append("text")
            .attr("x", label_coordinate.x)
            .attr("y", label_coordinate.y)
            .attr("text-anchor", "middle")
            .text(ft_name);
        }

        return basicChart;
    }


    private getBasicStandardChart(basicChart: Chart, config: Config): Chart {


       
        let xAxis = this.generateXAxis(config, basicChart);
        let yAxis = this.generateYAxis(config, basicChart);
        let y1Axis = this.generateY1Axis(config, basicChart);


        this.appendAxisLabels(config, basicChart)

        return basicChart;
    }

    getXAxisGenerator(config: Config): any {

        const height = config.chart.height - config.chart.margin.top - config.chart.margin.bottom;
        const width = config.chart.width - config.chart.margin.left - config.chart.margin.right;

        let xScale;
        let domainMin = config.axis.x.min;
        let domainMax = config.axis.x.max;

        if (config.chart.regression != null) {
            domainMin = config.data.points.reduce(function(prev: { x: number; }, curr: { x: number; }) {
                return prev.x < curr.x ? prev : curr;
            }).x;
            domainMax = config.data.points.reduce(function(prev: { x: number; }, curr: { x: number; }) {
                return prev.x > curr.x ? prev : curr;
            }).x;
        }

        if (!xScale) {
            if (config.chart.type === 'point') {
                xScale = d3
                .scaleLinear()
                .domain([domainMin, domainMax])
                .range([0, width]);
            } else if (config.chart.type === 'bar') {
                xScale = d3
                .scaleBand()
                .domain(config.axis.x.domain)
                .range([0, width]);
            } else {
                xScale = d3
                .scaleLinear()
                .domain([domainMin, domainMax])
                .range([0, width]);
            }
        }

        return {xAxisGenerator: d3.axisBottom(xScale as any), xScale: xScale};
    }


    generateXAxis(config: Config, basicChart: any): any {

        const height = config.chart.height - config.chart.margin.top - config.chart.margin.bottom;
        const width = config.chart.width - config.chart.margin.left - config.chart.margin.right;

        basicChart.height = height;
        basicChart.width = width;

        let {xAxisGenerator, xScale} = this.getXAxisGenerator(config);
        

        if (config.chart.showGrid) {
            const xTickSize = config.chart.height - config.chart.margin.top - config.chart.margin.bottom;
            const xAxisGrid = d3_axisBottom(xScale as any).tickSize(-xTickSize).tickFormat(x => '');
            basicChart.svg.append('g')
            .attr('class', 'y axis-grid')
            .attr('transform', 'translate(0,' + xTickSize + ')')
            .attr('color', '#ccc')
            .attr('opacity', 1)
            .call(xAxisGrid);
        }


                


        if (config.axis.x.tickCount != null) {
            xAxisGenerator.ticks(config.axis.x.tickCount);
        }


        let xAxis =  basicChart.svg.append("g")
              .call(xAxisGenerator);
        xAxis.attr("transform","translate(0," + height + ")");

        if (config.chart.fontSize) {
            xAxis.style('font-size', config.chart.fontSize+ 'px')
        }


        if (!basicChart['axis']) {
            basicChart['axis'] = {};
        }
        
        basicChart['axis']['x'] = {};
        basicChart['axis']['x']['axis'] = xAxis;
        basicChart['axis']['x']['scale'] = xScale;

        if (!config.axis.x.show) {
            xAxis.attr("display", "none")
        }

        // Dreht den Text auf der X-Achse
        // xAxis.selectAll("text")
        // .attr("transform", "translate(-10,10)rotate(-45)")
        // .style("text-anchor", "end")
        // .style("font-size", 20)
        // .style("fill", "#69a3b2")

        return xAxis;
    }

    private generateYAxis(config: Config, basicChart: any): any {
        const height = config.chart.height - config.chart.margin.top - config.chart.margin.bottom;
        const width = config.chart.width - config.chart.margin.left - config.chart.margin.right;

        let domainMin = config.axis.y.min;
        let domainMax = config.axis.y.max;

        if (config.chart.regression != null) {
            domainMin = config.data.points.reduce(function(prev: { y: number; }, curr: { y: number; }) {
                return prev.y < curr.y ? prev : curr;
            }).y;
            domainMax = config.data.points.reduce(function(prev: { y: number; }, curr: { y: number; }) {
                return prev.y > curr.y ? prev : curr;
            }).y;
        }

        let yScale = d3.scaleLinear()
               .domain([domainMin, domainMax])
               .range([height, 0]);

        
            if (config.chart.showGrid) {
                const yTickSize = config.chart.width - config.chart.margin.left - config.chart.margin.right;
                const yAxisGrid = d3_axisLeft(yScale as any).tickSize(-yTickSize).tickFormat(y => '');
                basicChart.svg.append('g')
                .attr('class', 'y axis-grid')
                .attr('color', '#ccc')
                .attr('opacity', 1)
                .call(yAxisGrid);
            }
        
        let yAxisGenerator = d3.axisLeft(yScale);

        if (config.axis.y.tickCount != null) {
            yAxisGenerator.ticks(config.axis.y.tickCount);
        }


        let yAxis =  basicChart.svg.append("g")
              .call(yAxisGenerator);

        if (config.chart.fontSize) {
            yAxis.style('font-size', config.chart.fontSize+ 'px')
        }


              
        if (!basicChart['axis']) {
            basicChart['axis'] = {};
        }
        basicChart['axis']['y'] = {};
        basicChart['axis']['y']['axis'] = yAxis;
        basicChart['axis']['y']['scale'] = yScale;

        if (!config.axis.y.show) {
            yAxis.attr("display", "none")
        }


        return yAxis;
    }

    private generateY1Axis(config: Config, basicChart: any): any {
        const height = config.chart.height - config.chart.margin.top - config.chart.margin.bottom;
        const width = config.chart.width - config.chart.margin.left - config.chart.margin.right;

        let yScale = d3.scaleLinear()
               .domain([config.axis.y1.min, config.axis.y1.max])
               .range([height, 0]);
        let yAxisGenerator = d3.axisRight(yScale);

        if (config.axis.y1.tickCount != null) {
            yAxisGenerator.ticks(config.axis.y1.tickCount);
        }


        let yAxis =  basicChart.svg.append("g")
              .call(yAxisGenerator);
        yAxis.attr("transform","translate(" + width + ",0)");
        if (!basicChart['axis']) {
            basicChart['axis'] = {};
        }
        basicChart['axis']['y1'] = {};
        basicChart['axis']['y1']['axis'] = yAxis;
        basicChart['axis']['y1']['scale'] = yScale;

        if (!config.axis.y1.show) {
            yAxis.attr("display", "none")
        }

        
        return yAxis;
    }

    private appendAxisLabels(config: Config, basicChart: any) {

        const height = config.chart.height - config.chart.margin.top - config.chart.margin.bottom;
        const width = config.chart.width - config.chart.margin.left - config.chart.margin.right;

        if (config.axis.x.label && config.axis.x.label.text) {
            

            let left = width;
            let top = height;

            if (config.axis.x.label.position === 'bottom-end') {
                top = top + config.chart.margin.top + config.chart.fontSize;
            } else if (config.axis.x.label.position === 'top-start') {
                left = config.chart.margin.left;
            } else if (config.axis.x.label.position === 'bottom-start') {
                left = config.chart.margin.left;
                top = top + config.chart.margin.top + config.chart.fontSize;
            } else if (config.axis.x.label.position === 'top-end') {
                left = left + config.chart.margin.left;
                top = top + config.chart.margin.top;
            }

            if (config.axis.x.label.marginLeft) { left = left + config.axis.x.label.marginLeft }
            if (config.axis.x.label.marginTop) { left = left + config.axis.x.label.marginTop }

            basicChart.svg.append("text")
                .attr("text-anchor", "end")
                .style("font-size", "34px")
                .attr("x", left)
                .attr("y", top)
                .text(config.axis.x.label.text);
        }



    }

    private angleToCoordinate(angle: any, value: any, radialScale: any, config: Config){
        let x = Math.cos(angle) * radialScale(value);
        let y = Math.sin(angle) * radialScale(value);
        return {"x": config.chart.width / 2 + x, "y": config.chart.height / 2 - y};
    }

    

    private generateBasicChartObject(): Chart {
        return {
            axis: {
                x: null,
                y: null,
                y1: null
            },
            svg: null,
            width: 0,
            height: 0
        }
    }

    getYAxisWidth(basicChart: Chart, config: Config) {
       if (!config.axis.y.show) { return 0; }
        let domainMin = config.axis.y.min;
        let domainMax = config.axis.y.max;
        const height = config.chart.height - config.chart.margin.top - config.chart.margin.bottom;

        let yScale = d3.scaleLinear()
               .domain([domainMin, domainMax])
               .range([height, 0]);

        let yAxisGenerator = d3.axisLeft(yScale);

        let width = this.getMaxLabelWidth(basicChart, yAxisGenerator, config);
        return width;
        

    }

    getXAxisHeight(basicChart: Chart, config: Config) {
        let {xAxisGenerator, xScale} = this.getXAxisGenerator(config);

        let width = this.getMaxLabelWidth(basicChart, xAxisGenerator, config);
        console.log(width);
        
        return Math.max(width, config.chart.fontSize * 1.5);

    }

    getMaxLabelWidth(basicChart: Chart, axisGen: any, config: Config) {
        let dummy = d3.select(basicChart.body).append('div');
        let svg = dummy.append("svg").style('overflow', 'visible').style('position', 'fixed').style('top', "166px").style('left', "561px").attr("class", "test123")
        let width = 0;
        let fontSize = config.chart.fontSize;
        svg.append('g').call(axisGen)
        .each(function () {
            d3.select(this).selectAll('text').each(function() {
                let box = this as SVGElement;
                let formattetText = box.innerHTML;
                let w = formattetText.length * fontSize / 2
                width = w > width ? w : width;
                
            })
        })

        dummy.remove();
        return width;
    }


}

